# Autogenerated by Astropy's setup.py on 2018-12-06 17:09:56 UTC
from __future__ import unicode_literals
import datetime

version = "3.1"
githash = "743055cada122d6da8704ad28452806690679ef5"


major = 3
minor = 1
bugfix = 0

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2018, 12, 6, 17, 9, 56)
debug = False

astropy_helpers_version = "3.1"

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
